from ._imports import *
from ._pconfig import *
from functools import cache
from dateutil.relativedelta import relativedelta

#TODO  currently life data source is using csv need to change to athena 
#TODO  salary success share currently using dummy data, also need to change to athena

def training_pipeline( runtime):   

    '''
    run_time = create_runtime ('inflow_premiums', 'dev', train_end=None)
    dict_results = training_pipeline(run_time)

    Parameters
    ----------
    runtime : dict
        dictionary contains parameters which will be used monitor pipeline.   

    Returns
    ------- 
    dict
        training results including tuned hyperparametersets, dataset used in training.
    ''' 

    cf_type = runtime['cf_type'] #'claim loss'
    train_begin = runtime['train_begin']   #'2020-05-01'
    train_end = runtime['train_end']  #None 
    test_len = runtime['test_len'] #8, 
    env= runtime['env'] #'dev' 

    #train_end is last month by default
    if not train_end:
        train_end = DtFind.first_day_last_month( dt.datetime.today(), dt_format=None).strftime("%Y-%m-%d")    

    df_origin, df_ts, dict_column_rename = get_train_data(cf_type,env,train_begin, train_end)  

    #get training range data
    train_df_prep, test_df_prep, df_selected_reset_prep = prep_prophet_training_data(df_ts, 
                                                                                    train_begin, 
                                                                                    train_end, 
                                                                                    test_len, 
                                                                                    dict_column_rename
                                                                                    )
    
    best_params, best_loss, winner, results = get_best_prophet_param_both2(train_df_prep, test_df_prep, test_len, cf_type)    
    #best_params['additional_feature_flag'] = best_params.get('additional_feature_flag', 0)  #optuna could skip this one
    dict_results = {
        'best params': best_params,
        'best loss': best_loss,
        'winner' : winner,
        'tuning results': results,
        'dataset inference': df_selected_reset_prep,  
        'original dataset':df_origin
        }
    return dict_results  


def get_train_data (cf_type, env, train_begin, train_end, generic_target_name='Values'):
    #mutual
    if cf_type == 'claim_loss':            
            df = get_claim_loss(env)
            df_ts = (df.pipe(add_mutual_month_number)
                                    .pipe(create_mutual_time_series,'total loss payments', 'Loss', train_begin, train_end, cf_type))
            dict_column_rename = {"year_month": "ds", "Loss": "y"}
    elif cf_type == 'claim_subrogation':
            df =  get_misc_adjust(env, filter='Claims Salvage and Subrogation')    
            df_ts = (df.pipe(add_mutual_month_number)
                                .pipe(create_mutual_time_series,'amount', 'sub_sal', train_begin, train_end, cf_type))
            dict_column_rename = {"year_month": "ds", "sub_sal": "y"}
    elif cf_type == 'vendor':
            df =  get_vendor_payments(env)
            df['amount'] = df['amount'].apply(lambda x : -x if x <0 else x)  #convert to positive value
            df_ts = (df.pipe(add_mutual_month_number)
                                .pipe(create_mutual_time_series,'amount', 'vendor', train_begin, train_end, cf_type))
            dict_column_rename = {"year_month": "ds", "vendor": "y"}
    elif cf_type == 'salary':
            df =  get_salary_payments(env)
            df['amount'] = df['amount'].apply(lambda x : -x if x <0 else x)  #convert to positive value    
            df['Date'] = pd.to_datetime(df['Date'])
            df_ts = df.loc[:, ['Date','amount','payday_count','success_share']].set_index('Date')
            dict_column_rename = {"Date": "ds", "amount": "y"}
    elif cf_type == 'inflow_premiums':
           df_invoice = prep_premium_inflow_invoice(env)
           premium_in = get_premium_inflow_history(env)
           data_with_invoice = premium_in.merge(df_invoice,left_on=['actual_month','actual_year'], right_on=['expected_clear_month','expected_clear_year'])
           #data_with_invoice['Date'] = pd.to_datetime(data_with_invoice['Date'])
           df = data_with_invoice.loc[:, ['Date','Premiums','invoice_due_amt']]
           #add additional feature
           #df_features = pd.read_csv(r"C:\Users\A036898\Downloads\a3539b06-967a-46f7-8506-8abc747b9aaf.csv")
           df_features  = get_additional_features(env)
           df_features['year_month'] = pd.to_datetime(df_features['year_month'])
           df = df.merge(df_features.rename(columns={'year_month':'Date'}), how = 'inner', on='Date').iloc[:,:2] #get original, remove invoice too            
           df_ts = df.set_index('Date')
           dict_column_rename = {"Date": "ds", "Premiums": "y"}
           save_additional_features_dataset (df_features, cf_type, env)
    elif cf_type == 'outflow_premiums':           
           df = get_premium_outflow_history(env).loc[:, ['Date','Dividend']]
           df['Dividend'] = df['Dividend'].apply(lambda x : -x if x <0 else x)  #convert to positive value
           #add additional feature from a query
           #df_features = pd.read_csv(r"C:\Users\A036898\Downloads\a3539b06-967a-46f7-8506-8abc747b9aaf.csv")
           df_features  = get_additional_features(env)
           df_features = df_features[['year_month', 'account_count', 'policy_count', 'total_premium',       
                            'unearned__premium_since_policy_start']]  #select highly correlated feature based on outflow reason
           df = df.merge(df_features.rename(columns={'year_month':'Date'}), how = 'inner', on='Date').iloc[:,:2]
           df['Date'] = pd.to_datetime(df['Date'])
           df_ts = df.set_index('Date')
           dict_column_rename = {"Date": "ds", "Dividend": "y"} 
           save_additional_features_dataset (df_features, cf_type, env)       
    
    elif cf_type =='inflow_receipts':  
           df_misc =  get_misc_adjust(env, 'Misc. Receipts')
           df_adj =  get_misc_adjust(env, 'Adjustments')
           df  =  pd.concat([df_misc, df_adj], ignore_index= True)
           df  =            (          #sort by year month number
                df.assign(mon_num = df['trans_month'].apply(lambda x: list(calendar.month_name).index(x)))
                  .sort_values(by = ['trans_year', 'mon_num'])
                  .drop(columns=['mon_num'])        
                )           
           df_ts = (df.pipe(add_mutual_month_number)
                    .pipe(create_mutual_time_series,'amount', 'misc_receipts', train_begin, train_end, cf_type))           
           df_ts['misc_receipts'] = cap_input_data(df_ts['misc_receipts'] )
           dict_column_rename = {"year_month": "ds", "misc_receipts": "y"}

    #life
    if cf_type[:4] == 'life':
        if cf_type == 'life_inflow_premiums':            
                df = get_life_inflow_premiums(env)
        elif cf_type == 'life_outflow_claims':            
                df = get_life_outflow_claims(env)
        elif cf_type in [ 'life_outflow_vendor', 'life_outflow_payroll', 'life_inflow_reinsurance'] :
                df = get_life_data(env, cf_type)
                if cf_type == 'life_outflow_payroll':  #salary model need additional features
                    df = get_life_salary_additional_features(df)                
        elif cf_type == 'life_receipts_adjustments':
                df = get_life_receipts_adjustments(env)
        df_ts = (df.pipe(add_life_month_number)
                                .pipe(create_life_time_series,generic_target_name, train_begin, train_end, cf_type))
        dict_column_rename = {"year_month": "ds", generic_target_name: "y"}     

    return df, df_ts, dict_column_rename


def get_claim_loss(env):
    query_claim_loss = 'SELECT * FROM silver_cashflow_forecast.claim_loss_summary'
    df_claim_loss = aws_athena_query('silver_cashflow_forecast', query_claim_loss, env)
    assert df_claim_loss.values.any(),  f"claim loss has missing value"
    df_claim_loss['total loss payments'] = df_claim_loss.apply(lambda r:r.loss_amt + r.exp_amt, axis =1)
    df_claim_loss = df_claim_loss[['trans_year','trans_month','total loss payments']]
    df_claim_loss['trans_month'] = df_claim_loss['trans_month'].str.strip()
    return df_claim_loss

def get_misc_adjust(env, filter='Claims Salvage and Subrogation'):
    query_sub = f'''SELECT * FROM 
            "silver_cashflow_forecast"."misc_receipt_adj"
            where cash_flow_typ = '{filter}'
            order by date
            '''
    df = aws_athena_query('silver_cashflow_forecast', query_sub, env)
    #make name consistent between dataset
    df['trans_year'] = df['date'].apply(pd.to_datetime).dt.year
    df['trans_month'] = df['date'].apply(pd.to_datetime).dt.month_name()
    df = df[['trans_year','trans_month','amount']]
    assert df.values.any(), f"{filter} has missing value"
    df['trans_month'] = df['trans_month'].str.strip()
    return df 




def add_mutual_month_number(df):
    df = df.copy()
    dict_month_mapping ={month: index for index, month in enumerate(calendar.month_name) if month}
    df['trans_month'] = df['trans_month'].str.strip().map(dict_month_mapping)
    df['trans_month_str'] = [ ("0" + str(i))[-2:]  for i in df['trans_month']]
    df['year_month'] = df['trans_year'].astype('str') + '-' +  df['trans_month_str'] 
    df = df.sort_values(by=['trans_year','trans_month'])
    return df 

def create_mutual_time_series(df, column, new_column_name, month_begin, month_end, cf_type):    
    df1 = df[['year_month', column]]
    df1.loc[:,'year_month'] = pd.to_datetime(df1['year_month'])
    df1.set_index('year_month', inplace=True)  #create time series index
    #month_begin, month_end =pd.to_datetime(month_begin), pd.to_datetime(month_end)
    df1 = df1.loc[month_begin:month_end]  #remove Apr 2024 outlier and last partial month
    if cf_type !='inflow_receipts':  #this type could have both postive and negative
        df1[column] = df1[column].apply(lambda x: -x if x< 0 else x)  #in athena table these are positive value, just in case
    #column_name = 'Loss' if column == 'total loss payments' else 'sal_sub' 
    df1.columns = [new_column_name]   #use in time series, short name
    return df1

#choose data range to use for training
def select_data_range (df_ts, begin, end=None):  
    '''
    select_data_range(df_claim_ts, begin = '2020-04-01')
    '''
    if not begin:
        begin=df_ts.index[0]

    if not end:
        end = df_ts.index[-1]

    return df_ts.loc[begin:end]

def prep_prophet_training_data (df, Begin_dt, end_dt, test_len,dict_column_rename ):
    '''
    This function is used to create train and test data for prophet hyperparameter tuning
    test_len is month number hold for test
    train_df_prep, test_df_prep = prep_prophet_training_data(df_subrogation_ts, '2020-05-01', '2024-09-01', 8, {"year_month": "ds", "Loss": "y"})
    '''
    df = df.copy()
    _df_selected =select_data_range(df, Begin_dt, end_dt)
    # if check_missing:
    #     missing_flag, monthlist = check_ts_missing_month(_df_selected, end_dt)
    #     assert not missing_flag, f'These month(s) data are missing {','.join(monthlist)}'    
    _df_selected_reset = _df_selected.reset_index().rename(columns=dict_column_rename)
    train_len = len(_df_selected_reset)- test_len
    _train_df = _df_selected_reset[0 : train_len]
    _test_df = _df_selected_reset[train_len : ]
    return _train_df, _test_df, _df_selected_reset


#########################add premium inflow code  by wcao
@cache
def get_clear_date(start_date, days_to_add):
    '''
    cache start date and clear date, save run time from 2 min to 2 sec
    '''
    current_date = pd.to_datetime(start_date)      
    while days_to_add > 0:
        current_date += dt.timedelta(days=1)
        if current_date.weekday() < 5 and (not us_holidays.get(current_date)):
            days_to_add -= 1
    return current_date


def calculate_expected_clear_date(row,dt_col='payment_due_dt',method_col='policy_payment_method'):
    business_days_to_add = {
    "ACH/EFT": 2.00,
    "Apply Back to Policy": 1.00,
    "AutoPay": 2.00,
    "AutoPay Bank": 2.00,
    "AutoPay Credit Card": 2.00,
    "Business Check/Money Order": 0.00,
    "Cash": 0.00,
    "Check": 0.00,
    "Credit Card": 2.00,
    "Foreign Check": 0.00,
    "Payment": 2.00,
    "Remote Scanner Deposit": 0.00
    }
    start_date = row[dt_col]
    days_to_add = business_days_to_add.get(row[method_col], 0)
    return get_clear_date(start_date, days_to_add)

def prep_premium_inflow_invoice(env):
    '''
    use lambda date to get the lastest record
    '''
    query_invoice = '''with payment_inflow_rnk as (
                select 
                *,
                row_number() over (partition by "rundate", "uw_company", "policy_payment_method", "payment_due_dt" order by "lambda_run_ts_utc" desc ) as rnk
                from silver_cashflow_forecast.payments_inflow_predictor
    )
            SELECT * FROM payment_inflow_rnk
            where payment_due_dt>date'2021-12-31'
            and rnk =1
    '''
    finserv_results = aws_athena_query('silver_cashflow_forecast', query_invoice, env)
    global us_holidays
    us_holidays = holidays.US()    
    finserv_results['expected_clear_date'] = finserv_results.apply(lambda x:calculate_expected_clear_date(x), axis=1).apply(pd.to_datetime)
    finserv_results['expected_clear_month'] = finserv_results['expected_clear_date'].dt.month
    finserv_results['expected_clear_year'] = finserv_results['expected_clear_date'].dt.year
    finserv_results_agg = finserv_results.groupby(['expected_clear_year','expected_clear_month']).agg({'invoice_due_amt':'sum'}).reset_index()
    return finserv_results_agg
	

def get_premium_inflow_history(env):  #prevent upstream duplicate
    query_premium = '''with cte_rnk as (
                SELECT year_month_day as "Date",year(year_month_day) as "Year",month(year_month_day) as "Month", payment_method,payment 
                    ,row_number() over (partition by year_month_day, payment_method, payment order by year_month_day) as rnk
                    FROM silver_cashflow_forecast.payments_inflow_1_month
                    where year_month_day>date'2021-12-31'
                )    
                select 
                "Date", 
                "Year", 
                "Month",
                payment_method,
                payment
                from cte_rnk where rnk =1'''
    premium_in = aws_athena_query('silver_cashflow_forecast', query_premium, env)
    premium_in['Date'] = premium_in.apply(lambda x: calculate_expected_clear_date(x,dt_col='Date',method_col='payment_method'),axis=1).apply(pd.to_datetime)
    premium_in['Date'] = pd.to_datetime(premium_in[['Year','Month']].assign(DAY=1))
    premium_in['Premiums'] = premium_in['payment'].astype('float')
    premium_in.dropna(axis = 0, inplace= True)
    premium_in['actual_month'] = premium_in['Date'].dt.month.astype(int)
    premium_in['actual_year'] = premium_in['Date'].dt.year.astype(int)
    premium_in_agg = premium_in.groupby(['actual_year','actual_month']).agg({'Premiums':'sum'}).reset_index()
    premium_in_agg['Date'] = pd.to_datetime(premium_in_agg[['actual_year', 'actual_month']].assign(day=1).rename(columns={'actual_year': 'year', 'actual_month': 'month'}))
    return premium_in_agg


def get_premium_outflow_history(env):
    query_premium = '''SELECT date as "Date", amount as Dividend, cash_flow_typ, adjustment_typ, filename, year, month, day   
	FROM "silver_cashflow_forecast"."misc_receipt_adj"
	WHERE cash_flow_typ like '%Premium/Dividend%'
	order by date'''
    premium_out = aws_athena_query('amica-eds-dev-datascience-db', query_premium, env)
    premium_out['Date'] = pd.to_datetime(premium_out['Date'],errors='coerce').apply(lambda x: x.replace(day=1))
    premium_out['Date'] = premium_out['Date'].dt.date
    premium_out['Date'] = premium_out['Date'].astype(str)
    return premium_out

def get_vendor_payments(env):
    query_vendor = '''SELECT date as "Date",Year(date) as trans_year,date_format(date, '%M') as trans_month, amount   
	FROM "silver_cashflow_forecast"."misc_receipt_adj"
	WHERE cash_flow_typ ='Vendor'
	order by date;'''
    df_vendor = aws_athena_query('silver_cashflow_forecast', query_vendor, env)
    assert df_vendor.values.any(),  f"vendor has missing value"
    df_vendor['Date'] = pd.to_datetime(df_vendor['Date'],errors='coerce').apply(lambda x: x.replace(day=1))
    df_vendor['Date'] = df_vendor['Date'].dt.date
    df_vendor['Date'] = df_vendor['Date'].astype(str)
    df_vendor['trans_month'] = df_vendor['trans_month'].str.strip()
    return df_vendor

def calculate_paydays(start_date,end_date,payday_weekday=4, start_payday=dt.datetime(2025,2,21)):
    paydays = []
    start_payday = dt.datetime(2025,2,21) #this represents a a
    current_date = start_date

    while current_date <= end_date:
        if current_date.weekday() == payday_weekday and (current_date - start_payday).days % 14 == 0: #Is Friday and biweekly
            paydays.append(current_date)
        current_date += relativedelta(days=1)

    monthly_counts = {}
    for payday in paydays:
        year = payday.year
        month = payday.strftime('%B')

        if (year, month) not in monthly_counts:
            monthly_counts[(year,month)] = 0
        monthly_counts[(year,month)] += 1
    
    data = [{'trans_year':year, 'trans_month':month, 'payday_count':count} for (year,month), count in monthly_counts.items()]
    data = pd.DataFrame(data)
    return data

def get_payday_datataframe():
    
    current_date = dt.datetime.now()
    start_date = dt.datetime(2020,1,1)
    end_date = current_date + relativedelta(months=4)
    end_date = end_date.replace(day=1) + relativedelta(days=-1)
    
    df = calculate_paydays(start_date,end_date)
    df['success_share'] = np.where(df['trans_month']=='February',1,0)
    return df

def get_salary_payments(env):
    query_salary = '''SELECT date as "Date",Year(date) as trans_year,date_format(date, '%M') as trans_month, amount   
	FROM "silver_cashflow_forecast"."misc_receipt_adj"
	WHERE cash_flow_typ ='Salary and Benefits'
	order by date;'''
    df_salary = aws_athena_query('silver_cashflow_forecast', query_salary, env)
    payday_df = get_payday_datataframe()
    
    #This needs to be replaced with misc_adjustment table^M
    success_sharing_ath = aws_athena_query('silver_cashflow_forecast',
                                       '''select * FROM "silver_cashflow_forecast"."misc_receipt_adj" where cash_flow_typ = 'Success Share';''',env)    
    #success_sharing_ath['date'] = pd.to_datetime(success_sharing_ath['date']).dt.to_period('M').dt.to_timestamp().astype(str)^M    
    df_salary = df_salary.merge(payday_df,on=['trans_year','trans_month'])[['Date','trans_year','trans_month','amount','payday_count','success_share']]
    assert df_salary.values.any(),  f"salary has missing value"
    df_salary['Date'] = pd.to_datetime(df_salary['Date'])+pd.offsets.MonthEnd(0)
    success_sharing_ath['date'] = pd.to_datetime(success_sharing_ath['date'])+pd.offsets.MonthEnd(0)
    for i,row in success_sharing_ath.iterrows():          
          df_salary['amount'] = np.where(df_salary['Date']==row['date'], df_salary['amount']-row['amount'], df_salary['amount'])
    df_salary['Date'] = pd.to_datetime(df_salary['Date'],errors='coerce').apply(lambda x: x.replace(day=1))
    df_salary['Date'] = df_salary['Date'].dt.date    
    df_salary['trans_month'] = df_salary['trans_month'].str.strip()
    return df_salary

#########################add misc receipts code  by Wcao
def cap_input_data (s:pd.Series):
        mean = s.mean()
        std_dev = s.std()
        upper_bound = mean + 2 * std_dev
        lower_bound = mean - 2 * std_dev

        s = np.where(s>upper_bound,upper_bound, 
                                        np.where(s< lower_bound,lower_bound,s))
        s = np.where(s==0,mean,s)
        return s


########################## additional features by athena
def get_additional_features(env):        
    if env ==  'dev':
        db = 's3tablescatalog_cda_iceberg_pc_perf_merge_public_link'
    elif env =='prod':
        db = 's3tablescatalog_cda_iceberg_pc_prod_merge_public_link'
    query = '''
    with date_series AS (
    SELECT
        CAST(t.date AS DATE) AS full_date
    FROM
        UNNEST(SEQUENCE(DATE '2020-04-01', date_add('month', 3, current_date), INTERVAL '1' DAY)) AS t(date)
    ),
    
    date_range as (
    SELECT
        DATE_TRUNC('month', full_date) AS datebegin,
            date_trunc('month',  full_date) + interval '1' month - interval '1' day  AS dateend
    FROM
        date_series
    GROUP BY
        DATE_TRUNC('month', full_date)
    -- ORDER BY
    --     month_begin_date
        )    ,
    temp_policy as (
    select 
            distinct
            d.datebegin,
            c.AccountNumber,
            ((SUBSTRING(a.policynumber,1,1) || SUBSTRING(a.policynumber,5,9))) as PolicyNumber, 
            PeriodStart, 
            PeriodEnd, 
            a.TotalCostRPT,
            a.TotalCostRPT/DATE_DIFF('day', PeriodStart, PeriodEnd) AS DailyPremium,
            DATE_DIFF('day',
                CASE WHEN PeriodStart > d.datebegin THEN PeriodStart ELSE d.datebegin END,
                CASE WHEN PeriodEnd < d.dateend THEN PeriodEnd ELSE d.dateend END) + 1 AS DaysInMonthOverlap,
            DATE_DIFF('day',
                PeriodStart,
                CASE WHEN PeriodEnd < d.dateend THEN PeriodEnd ELSE d.dateend END) + 1 AS accum_days
            from pc_policyperiod a 
                join pc_policy b on a.PolicyID = b.ID 
                join pc_account c  on b.AccountID = c.ID
                join date_range d  on 1=1
            where 
            a.MostRecentModel =true  
            and TotalCostRPT > 0
            and cast(a.Periodstart as date) <= d.dateend AND cast(a.PeriodEnd as date) >= d.datebegin
    )

    select 
    --date_format(datebegin, '%Y-%m') as year_month,
    datebegin as year_month, 
    count(distinct AccountNumber) as account_count,
    count(distinct PolicyNumber) as policy_count,
    sum(TotalCostRPT ) as total_premium,   --given inflight policy all premium
    sum(DaysInMonthOverlap * DailyPremium ) as earned_premium_this_month,
    sum(accum_days * DailyPremium) as earned_premium_since_policy_start,
    sum(TotalCostRPT ) - sum(accum_days * DailyPremium) as unearned__premium_since_policy_start
    from temp_policy
    group by datebegin
    order by datebegin
    '''
    df_features =aws_athena_query(db, query, env)
    #save the additional feature for future reference 

    return df_features
    



###########################################################
#life inflow premium

def get_life_inflow_premiums(env):
    #TODO use athena query
    df_inflow_premiums  = pd.read_csv(r"\\hofile01\Enterprise Data Solutions\EDS Initiatives\Cash Flow Forecasting\Cashflow for life\Premiums\Data\cashflow_life_premium_txns_agg.csv" )
    df_inflow_premiums  = df_inflow_premiums[['Processed_Year', 'Processed_Month','Tre_AMTPROCESSED']]
    return df_inflow_premiums 

# life outlfow claims
def get_life_outflow_claims(env):
    #TODO use athena query
    df_outflow_claims  = pd.read_csv(r"\\hofile01\Enterprise Data Solutions\EDS Initiatives\Cash Flow Forecasting\Cashflow for life\Claims\Data\cashflow_life_claims_txns_agg.csv" ) 
    df_outflow_claims  = df_outflow_claims[['Processed_Year', 'Processed_Month','Tre_AMTDISBURSED']]
    return df_outflow_claims 


def get_life_receipts_adjustments(env):
    df_receipts = get_life_data(env, cf_type = 'life_inflow_misc_receipts')
    df_adjustments = get_life_data(env, cf_type = 'life_outflow_adjustments')
    df_receipts_adjustments = pd.concat([df_receipts, df_adjustments], ignore_index=True).sort_values(by=['Processed_Year', 'Processed_Month']).reset_index(drop=True)
    return df_receipts_adjustments


def get_life_data(env, cf_type = 'life_outflow_vendor'):
    ''''
    get data from life athena table, need to change hard code csv    '''
    
    dict_cf_type = {'life_outflow_vendor':'Vendor',
                    'life_inflow_misc_receipts':'Misc. Receipts',
                    'life_outflow_adjustments':'Adjustments',
                    'life_outflow_payroll':'Salary and Benefits',
                    'life_inflow_reinsurance':'Reinsurance'                  
                    } 
    #df = pd.read_csv(r"\\hofile01\Enterprise Data Solutions\EDS Initiatives\Cash Flow Forecasting\Cashflow for life\Misc Receipts_Vendor_Salary and Benefits\misc_receipts_vendor_salary_benefits.csv")
    df = pd.read_csv(r"\\hofile01\Enterprise Data Solutions\EDS Initiatives\Cash Flow Forecasting\Cashflow for life\Actuals\Cashflow_life_actuals.csv")
    df['Date'] = pd.to_datetime(df['Date'] )
    df= df[df['Cash_flow_type']==dict_cf_type[cf_type]].sort_values(by= 'Date')
    df =  df[df['Amount'].str.strip()!='-']  #remove - row
    df['Processed_Year'] = df['Date'].dt.year
    df['Processed_Month'] = df['Date'].dt.month
    df ['Amount'] = [ -1*float(i) if (float(i)<0) and dict_cf_type[cf_type] not in ['Misc. Receipts', 'Adjustments'] else float(i)  for i in df ['Amount'].apply(lambda x: x.replace(',', '')) ]  
    df = df [['Processed_Year', 'Processed_Month', 'Amount' ]]
    return df



#general function

def add_life_month_number(df):
    df = df.copy()
    dict_rename = dict(zip(df.columns[:2], ['Year', 'Month']))
    if str(df.iloc[:, 1].dtype)[:3]  == 'int':  #check if month column is number     
        df['year_month'] = pd.to_datetime(df.rename(columns=dict_rename)[['Year', 'Month']].assign(DAY=1))
        
    else: 
        raise ValueError ('convert month to number in the previous step')
    return df

def create_life_time_series(df, column, month_begin, month_end, cf_type):     #here name predict column generic
    df = df.copy()
    df.set_index('year_month', inplace=True, drop=True)  #create time series index
    df = df.iloc[:, 2:]  #get the predict column, remove year month
    df.columns = [column if i==0 else c for i, c in  enumerate(df.columns)]  #give generic name
    #month_begin, month_end =pd.to_datetime(month_begin), pd.to_datetime(month_end)
    df = df.loc[month_begin:month_end]  #remove Apr 2024 outlier and last partial month
    ### convert negative value, modify in life
    if cf_type !='life_receipts_adjustments':  #this type could have both postive and negative
        df[column] = df[column].apply(lambda x: -x if x< 0 else x)  #in athena table these are positive value, just in case     
       #use in time series, short name
    return df

def get_life_salary_additional_features(df_salary):
    payday_df = get_payday_datataframe()
    payday_df['trans_month'] =  pd.to_datetime(payday_df['trans_month'], format='%B').dt.month  #convert to month number
    df_salary = (df_salary.merge(payday_df, how = 'left', left_on=['Processed_Year','Processed_Month'], right_on = ['trans_year','trans_month'])
            .drop(['trans_year','trans_month'], axis=1)   #remove left side join column
            )
    #TODO this is dummy data, need to replace with real data
    dict_success_share = {
        'date': [ '2022-02-28', '2023-02-28', '2024-02-28', '2025-02-28'],  #will this automatically loaded?
        'amount':[1219213.75,
                1010019.08,
                3199405.16,
                4295804.42,
            ]
        }
    df_success_share = pd.DataFrame(dict_success_share)
    #add join column
    df_success_share['date'] = pd.to_datetime(df_success_share['date'])
    df_success_share['year'] = df_success_share['date'].dt.year
    df_success_share['month'] = df_success_share['date'].dt.month
    #rename column
    df_success_share.columns =['date', 'deduct_amount', 'year', 'month']  

    #merge two dataset and deduct the amt
    df_salary = df_salary.merge(df_success_share, how = 'left', left_on = ['Processed_Year', 'Processed_Month'], right_on = ['year', 'month'])
    df_salary['Amount'] = df_salary.apply(lambda x: x['Amount'] if pd.isna(x['deduct_amount']) else x['Amount'] - x['deduct_amount'], axis =1)
    df_salary = df_salary.drop(['date',	'deduct_amount', 'year','month'], axis =1)
    return df_salary 

def save_additional_features_dataset (df, cf_type, env):
    '''
    Save earlist additional feature dataset for troubleshooting 
    Parameters
    ----------
    df : DataFrame 
                The additonal feature dataframe.
    cf_type : str 
                cashflow type.
    env : str 
                environment you are using.
    Examples
    ----------
    save_additional_features_dataset (df_dummy, cf_type, env)
    '''
    datestamp = dt.datetime.now().strftime("%Y%m")
    s3_file_path = os.path.join('cash_flow_forecasting/model_misc/additional_data_hx/',f'cash_flow_model_{cf_type}_additional_features_{datestamp}.csv' )
    if s3_check_file_exists(s3_file_path, env):
        pass  #if file exist, not save
    else :  s3_save_and_upload_file(
            df,
            s3_file_path = s3_file_path,
            overwrite=False,
            env=env,
        )
    return 


# def prep_prophet_training_data (df, Begin_dt, end_dt, test_len,dict_column_rename ):
#     '''
#     This function is used to create train and test data for prophet hyperparameter tuning
#     test_len is month number hold for test
#     train_df_prep, test_df_prep = prep_prophet_training_data(df_subrogation_ts, '2020-05-01', '2024-09-01', 8, {"year_month": "ds", "Loss": "y"})
#     '''
#     df = df.copy()
#     _df_selected =select_data_range(df, Begin_dt, end_dt)    
#     _df_selected_reset = _df_selected.reset_index().rename(columns=dict_column_rename)
#     train_len = len(_df_selected_reset)- test_len
#     _train_df = _df_selected_reset[0 : train_len]
#     _test_df = _df_selected_reset[train_len : ]
#     return _train_df, _test_df, _df_selected_reset
    



__all__ = [
    'get_ai_subro_possiblity', 'extract_reason_possibility', 'show_record_detail_by_index'
    
    ]

from ._imports import *
from botocore.exceptions import ClientError
from functools import cache



class Aws_Bedrock():
    # aws_cred = {

    #     ##currently we don't have permanent credential, need to replace everytime
    #     "region_name" : 'us-east-2',
    #     "aws_access_key_id"    : "ASIAVWZBVNN4UJN46TBL",
    #     "aws_secret_access_key"   : "sLBCnUU31wQi6YqsDPa+Z8qYGbrk5IhdCshsP2dD",
    #     "aws_session_token"       : "IQoJb3JpZ2luX2VjEKX//////////wEaCXVzLWVhc3QtMSJHMEUCIHJCvPXb8PHxjQhJt91MW4Xd6i1TlOFC5pMFiXWP6BwIAiEAxGZ5aQFBSdCjc6JoFHd1PiK3b22qr+tb3VIt5VtwTdIqgwMIrv//////////ARABGgwzOTI1MjMyNDY0NTciDJ22TrsB8tzfDxM8jSrXArr/deZxcIDUWVnGRZ3UOzaNHf9vZA6Sr/0y9KPP9jOd2dAzBVPcW4MS4PgZOE/3jKFVVke09fgx0EOzGH0qtoo1iAQOQ3uKsvFNQtLH3/N2PmL53GTHPrp+W5uM86fuYbKg6Q7aqc3+fktmMBiNGcOaIGbdCoKAaemK17FYtgV9bFkhoqcdhGAs2ax98hXnYsHK7fc0f9Wb6wKIZFZfYEof4Etf9JqWM3sNMzXPg72sKd8X3un7EwoIaAolaX6Bwn9WExj3Vapn8rS553zGHG7Te4VuDCS7krQyDtXIjnziQU6xeiyk21cTvbPPt0b2DgoPIuriuoiYUZTTQvVttQduDFOGIAX3YHPFBlHiUXO1KL+inxCoQTpRBpm9ZBS4A+dlv/Zs00gqbccPnwFZSi1OlKiP/V86NaPUq0fz5c1ZoWQYpik3Vk+nHSXRNYHLibFI9lp1vLcwia67wwY6pwEFGKq6kQkk1JpAhBtQ6hVDwluczF+8LWFrmXPSdNIRlWlZ8P0zdK7J+T/lvv2HUB95NuyHCh9yfsi2qzhuSloIRdsSXyyEWCuGfxZy31sfpK64/+sjLZnJ8YchDQt62aU67aixcWkvpqMAvdx67qiKYgmP/JKl+r2gvCb/CQqEn94YGqmgcJyM5NaDYo7/l1VR+eiqeks9WqWq33nb3PSdifVjxmpa8g=="

    # }

    # bedrock_client =  boto3.client(service_name="bedrock", **aws_cred )
    # bedrock_runtime_client = boto3.client(service_name= "bedrock-runtime", **aws_cred )     
     
    # use SSO
    session = boto3.Session(profile_name='EDS-Scientists-Dev-392523246457')
    bedrock_client =  session.client(service_name="bedrock")
    bedrock_runtime_client = session.client(service_name= "bedrock-runtime")    


    @staticmethod    
    def list_foundation_models():
        """
        Gets a list of available Amazon Bedrock foundation models.

        :return: The list of available bedrock foundation models.
        """

        try:
            response = Aws_Bedrock.bedrock_client.list_foundation_models()
            models = response["modelSummaries"]
            print("Got %s foundation models.", len(models))
            return models

        except ClientError:
            print("Couldn't list foundation models.")
            raise


    #https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/bedrock-runtime/client/invoke_model.html
    @staticmethod
    def invoke_bedrock_model(system_prompt, user_prompt, region_name="us-east-2"):
        """
        Invokes a Bedrock model and returns the generated text.

        Args:
            model_id (str): The ID of the Bedrock model to invoke.
            prompt (str): The input prompt for the model.
            region_name (str, optional): The AWS region. Defaults to "us-east-1".

        Returns:
            str: The generated text from the model, or None if an error occurred.
        """
        # try:        

        # Prepare the input payload            

        native_request = {
        "anthropic_version": "bedrock-2023-05-31",
        "system": system_prompt,
        "max_tokens": 4096,
        "temperature": 0.2,
        "messages": [
            {
                "role": "user",
                "content": [{"type": "text", "text": user_prompt}],
            }
        ],
        }

        response = Aws_Bedrock.bedrock_runtime_client.invoke_model(
            modelId=r"arn:aws:bedrock:us-east-2:392523246457:inference-profile/us.anthropic.claude-3-7-sonnet-20250219-v1:0",#model_id,
            contentType="application/json",
            accept="application/json",
            body=json.dumps(native_request),
            guardrailIdentifier= r'arn:aws:bedrock:us-east-2:392523246457:guardrail/wof2sh6wm93z',
            #X-Amzn-Bedrock-GuardrailIdentifier = r'arn:aws:bedrock:us-east-2:392523246457:guardrail/wof2sh6wm93z',
            guardrailVersion='6'
            
        )
        
        model_response = json.loads(response["body"].read())
        model_text = model_response["content"][0]["text"]
        return model_text

@cache
def get_ai_subro_possiblity (cleaned_fnol, fault_rating ):
    if not cleaned_fnol:
        return None
    system_prompt = '''
                    You are a claim subrogation expert. Insurance company needs to recoup the money when they found the third party is liable.        
        you are presented with loss notes and try to come up with subrogation probability. In the notes, the first party can be called as insured, insd, A, etc,  whereas the third party can be called as claimant, clmt, C, etc.
         I would like you tell me if the third party is responsible for the loss, if the third party does, output a list like [reason, the percentage from 1 to 100%]. if the first party is responsible for the loss, output[ reason, 0%],
        please limit reason in 10 words, only output the list please
    '''
    user_prompt = f'''I have a column,  contain the notes, I would like you tell me if the third party is responsible for the loss, if the party does, output a list like [reason, the percentage from 1 to 100%]. if not, output[ reason, 0%], 
         please apply the following rule:
            1. if {fault_rating} is "Insured at fault", that means third party is not liable.
            2. if third party is hit and run  the case usually not consider third party is liable.
            3. if the incident cannot find the third party, the case is not liable, senario like Rock from Road, Falling/Flying Objects, Malicious Mischief and Vandalism, Weather Related (ice storm, hurricane, etc.)
            4. if the case is animal claim, usually that is not liable, but farm animal escape from enclosure, the owner is liable.
         here is the note: {cleaned_fnol}. only output the list please'''
    generated_text = Aws_Bedrock.invoke_bedrock_model(system_prompt, user_prompt)
    return generated_text


def try_ai_prediction_multiple_time(df, dest_column = 'ai_result', src_column ='loss_note_cleaned',  counterlimit = 5, func =get_ai_subro_possiblity ):    
    '''
    OpenAI API could skip record randomly, usually less than 1%, this will ensure skip record is reevaluated
    might not use for bedrock
    '''
    df[dest_column] = None
    counter = 0
    while df[dest_column].isna().sum() >0 and counter <counterlimit:
        df_ai_result_missing = df[df[dest_column].isna()]
        for i in df_ai_result_missing.index:
            df.at[i, dest_column] = func(df.at[i, src_column], df.at[i, 'fault_rating'])
        counter +=1
        print(counter, df[dest_column].isna().sum())
        func.cache_clear()  #clear cache to evaluate freshly
    return


def extract_reason_possibility (text):
    '''extract result from AI output'''    
    pattern = r'\[(.*), (\d{1,3}%)]' 
    match = re.search(pattern, text, re.IGNORECASE)
    if not match:
            return None, None   
    else:
        reason, possiblity = match.group(1), match.group(2)
        return reason, possiblity
    


def show_record_detail_by_index(df, ind):
    '''show detail info for given index for troubleshooting '''
    if 'loss_note' in df.columns:
        print(df.at[ind, 'loss_note' ])
    if 'fnol' in df.columns:
        print(df.at[ind, 'fnol' ])
    if 'claimnumber' in df.columns:
        print(df.at[ind, 'claimnumber' ])
    if 'fault_rating' in df.columns:
        print(df.at[ind, 'fault_rating' ])
    return 

@cache
def get_ai_police_report_subro_possiblity (police_report ):
    '''
    df_police['police_report_AI_result'] = df_police['content_from_pdf'].apply(get_ai_police_report_subro_possiblity)
    '''
    if not police_report:
        return None
    system_prompt = '''
                    You are a claim subrogation expert at Amica. Insurance company needs to recoup the money when they found the third party is liable.        
        you are presented with a police report and try to find if policeman state the third party is at fault. 
    '''
    user_prompt = f'''The police report will ususally contain a section where insurce companies are identified. for example, "Insurance Company  [2] 'AMICA'" means second vehicle/driver's insurance is with Amica in the report. if you don't find such info, please identify amica related vehicle number from context. 
    please note, from subrogation perspective, Amica insured is first party, the other vehicles are the third party.  please read the police report content and  
    tell me if the police stated the third party is responsible for the loss.  If the party does, output a list like [the number related with Amica vehicle, reason, the percentage from 1 to 100%]. if not, output[ the number related with Amica vehicle, reason, 0%], If police did not state the fault clearly, please output [the number related with Amica vehicle, "insufficient information to determine responsibility", 50%].
    please limit reason in 10 words, only output the list please. Please use policeman's judgement, not your own judgement.
    the police report is as follows {police_report}
              '''
    generated_text = Aws_Bedrock.invoke_bedrock_model(system_prompt, user_prompt)
    return generated_text

def extract_police_report_reason_possibility (text):
    '''extract result from AI output
    df_police[['vehicle_no', 'police report reason', 'police report subro possiblity']] = df_police.apply(lambda x: extract_police_report_reason_possibility(x['police_report_AI_result']), axis =1, result_type='expand' )
    '''    
    pattern = r'\[(\d), (.*), (\d{1,3}%)]' 
    match = re.search(pattern, text, re.IGNORECASE)
    if not match:
            return None, None, None   
    else:
        vehicle_no, reason, possiblity = match.group(1), match.group(2), match.group(3) 
        return vehicle_no, reason, possiblity


from ._hail import *
from ._imports import *
from ._pconfig import pconfig

def get_hail_storm_polygon(cat_no, cat_date, save_gdf = False):
    hd = HailData(cat_date = cat_date)
    data = hd.get_hail_data()
    df, gdf = hd.extract_data_to_df(data)
    if save_gdf:
        path = pconfig['gis_file_folder']
        hd.save_data_csv(df, gdf, path= path )
        print(f"gdf have saved at {path}\hail_swath_{hd.file_timestamp}.gpkg")
    return gdf

def get_cat_geodataframe (gdf_cat_polygon, df_claim, df_policy):

    #create claim gdf
    geometry = [Point(xy) for xy in zip(df_claim['GEO_X'], df_claim['GEO_Y'])]
    # Convert to GeoDataFrame
    gdf_claim = gpd.GeoDataFrame(df_claim, geometry=geometry, crs="EPSG:4326") 

    #has claim in the swath    
    gdf_has_claim_inside_swath = gpd.sjoin(gdf_claim, gdf_cat_polygon, how="inner", predicate="within")  #now sure about join condition
    #gdf_has_claim_inside_swath[['CAT_NO', 'POLICY_NO', 'FILE_NO','LOSS_DT','diameter_in', 'diameter_in_label']]


    #has claim not in the hail swath
    gdf_has_claim_not_in_swath =  gdf_claim[~gdf_claim['POLICY_NO'].isin(gdf_has_claim_inside_swath['POLICY_NO'])]
    
    
    #create policy gdf
    geometry_h = [Point(xy) for xy in zip(df_policy['GEO_X'], df_policy['GEO_Y'])]
    # Convert to GeoDataFrame
    gdf_policy = gpd.GeoDataFrame(df_policy, geometry=geometry_h, crs="EPSG:4326") 
    

    #this is all policy inside swath
    joined_inner_policy = gpd.sjoin(gdf_policy, gdf_cat_polygon, how="inner", predicate="within")  #now sure about join condition
    
    #this is all policy inside swath but not have claim 
    gdf_has_no_claim_in_hailswath= joined_inner_policy[~joined_inner_policy['POLICY_NO'].isin(df_claim['POLICY_NO'].unique())]
    
    return gdf_has_claim_inside_swath, gdf_has_claim_not_in_swath, gdf_has_no_claim_in_hailswath 
    

#get distance and nearest polygon for all gdf has claim but not in swath

def convert_gdf_epsg (gdf,  target_epsg = 3857):
    gdf =  gdf.copy()
    #epsg_code = gdf.crs.to_epsg()
    #print(f'current epsg code is {epsg_code}')
    gdf = gdf.to_crs(epsg=target_epsg)
    return gdf

def get_nearest_polygon(query_point_geom, gdf):
    '''
    gdf is hail polygon from hail api, by default is crs 4326,need to convert to 3857 for distance calculation     
    query_point_geom from a gdf also need to convert to 3857
    '''
    gdf_dis = convert_gdf_epsg(gdf, 3857)    
    distances = gdf_dis.distance(query_point_geom)
    # Find the index of the minimum distance
    closest_polygon_index = distances.idxmin()
    # Retrieve the closest polygon as a GeoSeries or GeoDataFrame row
    closest_polygon = gdf_dis.loc[closest_polygon_index]
    return closest_polygon['geometry']

def get_nearest_distance_polygon(gdf_claim_not_in_swath, gdf_polygon):

    gdf_claim_not_in_swath_dis = convert_gdf_epsg(gdf_claim_not_in_swath, target_epsg = 3857)
    gdf_dis = convert_gdf_epsg(gdf_polygon, 3857)     
    nearest_geoms = gpd.sjoin_nearest(
        gdf_claim_not_in_swath_dis, #points_gdf, 
        gdf_dis ,#polygons_gdf, 
        how="left", 
        distance_col='distance_to_nearest_poly'  #"dist_to_polygon"
    )
    nearest_geoms.drop_duplicates(subset=['POLICY_NO'], keep='first', inplace=True)  #some show multiple row with distance as 0, possible at the edge?
    gdf_claim_not_in_swath['distance'] = nearest_geoms['distance_to_nearest_poly'] 
    gdf_claim_not_in_swath['nearest_polygon_diameter_in'] = nearest_geoms['diameter_in']
    gdf_claim_not_in_swath['nearest_polygon'] = gdf_claim_not_in_swath['geometry'].apply(lambda x: get_nearest_polygon(x,gdf_dis) )
    return gdf_claim_not_in_swath


from ._imports import *
from ._pconfig import *
from ._score import *
from ._misc import get_model_version
from ._train import get_train_data
from statsmodels.tsa.seasonal import STL

def monitor_pipeline (runtime):
    '''
    Example
    -------
    cf_type = 'life_inflow_premiums'
    #train_begin = None   
    train_end = None 
    env = 'dev'  
    runtime = create_runtime (cf_type, env, train_end=None)
    tracker = monitor_pipeline (runtime)
    print(tracker)

    Parameters
    ----------
    runtime : dict
        dictionary contains parameters which will be used monitor pipeline.   

    Returns
    ------- 
    dict
        monitor results.

    '''
    df_prev = get_prev_data(runtime)
    df_curr = get_curr_data(runtime) 

    df_prev_compare, df_curr_compare = adjust_data_per_features_used(df_prev, df_curr)    
    data_drift_flag, data_drift_detail = get_data_drift(df_prev_compare, df_curr_compare)    
    outlier_flag, outlier_detail = get_outlier(df_curr, runtime)
    null_flag, missing_months = check_ts_missing_month(df_curr, runtime['train_end'])
    tracker ={
        'Data Drift' :data_drift_flag,
        'Data Drift Detail': data_drift_detail,
        'Outlier Drift': outlier_flag,
        'Outlier Detail': outlier_detail, 
        'Null Drift': null_flag,
        'Missing Months':missing_months,
        'Prev Dataset': convert_ts_to_dict(df_prev),
        'Curr Dataset': convert_ts_to_dict(df_curr)
    }
    return tracker

def convert_series_to_df (df):
    '''change to df if it is series type'''
    df = df.copy()     
    if isinstance(df , pd.Series):
                df  = df.to_frame()
    return df

def get_data_with_ts(df, runtime):
    '''
    add timeseries to the index of existing df    
    '''
    df_ind = df.copy()
    df_ind.index =df_ind.iloc[:, :2].apply(lambda x: str(x[0]) + ' ' + str(x[1]), axis =1).apply(pd.to_datetime).values  #first two column are year month
    df_ind = df_ind.loc[df_ind.index <= runtime['train_end']].iloc[:, 2:] #what if there are mulitple feature iloc[:, 2:]?
    df_ind = convert_series_to_df(df_ind)
    return df_ind

def pre_process_mutual_data(df, runtime):
    df = df.copy( )
    if 'Date' in df.columns:
        df['Date'] = pd.to_datetime(df['Date'])
    else:
        df['Date'] = pd.to_datetime(df['trans_year'].astype(str) + '-' + df['trans_month'] +'-01')
    extra_columns = [ i for i in df.columns if i in ['trans_year', 'trans_month']]
    for i in extra_columns:
        df.drop(i, axis =1, inplace = True)                
    df = df[df['Date']<=pd.to_datetime(runtime['train_end'])].set_index('Date') 
    return df      

def get_prev_data(runtime):
    '''
    based on version to get reference file, make date as index
    '''
    model_version = get_model_version(runtime)
    if not model_version:  #first time run
         df_prev = get_curr_data(runtime)
    else:         
        file_list = s3_list_folder_contents(pconfig[f's3_model_assets_folder_{runtime['env']}'], None, f'^.*cash_flow_model_{runtime['cf_type']}_{model_version}.csv$', runtime['env'])
        latest_file = sorted(file_list)[-1]
        df_prev = s3_download_and_read_csv(latest_file, env = runtime['env'])
        #mutal side
        if runtime['cf_type'] in ['claim_loss', 'claim_subrogation', 'inflow_premiums', 'outflow_premiums', 'inflow_receipts', 'vendor', 'salary'] :            
            df_prev = pre_process_mutual_data(df_prev, runtime)
        else:
            df_prev = get_data_with_ts(df_prev, runtime)
        
    return df_prev

def get_curr_data(runtime):
    type, env, train_begin, train_end= runtime['cf_type'], runtime['env'], runtime['train_begin'], runtime['train_end']
    df_curr, *_ = get_train_data (type, env, train_begin, train_end)
    if runtime['cf_type'] in [ 'inflow_premiums', 'outflow_premiums', 'claim loss', 'claim subrogation','inflow_receipts', 'vendor', 'salary'] :            
            df_curr= pre_process_mutual_data(df_curr, runtime)
    else:
        df_curr = get_data_with_ts(df_curr, runtime)
    return df_curr

def adjust_data_per_features_used(df_prev, df_curr):
    shared_columns =  [ i  for i in df_prev.columns if i in df_curr.columns]
    df_prev_compare = df_prev[shared_columns]
    df_curr_compare = df_curr[shared_columns]
    return df_prev_compare, df_curr_compare 


def chi_square(ref, new, alpha=0.05):
    categories     = set(ref.unique()).union(set(new.unique()))
    ref_proportion = [(ref==c).sum()/len(ref) for c in categories]
    new_proportion = [(new==c).sum()/len(new) for c in categories]
    r = scipy.stats.chisquare(new_proportion, f_exp=ref_proportion)
    return int(r.pvalue < alpha)


def population_stability_index(ref, new, buckets=10, psi_cutoff=0.2):
    '''
    Calculate the PSI (population stability index)
    Adapted from https://github.com/mwburke/population-stability-index/blob/master/psi.py
    '''

    def sub_psi(e_pct, a_pct):
        if a_pct == 0: a_pct = 0.0001
        if e_pct == 0: e_pct = 0.0001
        return (e_pct - a_pct) * np.log(e_pct / a_pct)

    breaks  = np.linspace(np.min(ref), np.max(ref), buckets+1)
    ref_pct = np.histogram(ref, breaks)[0] / len(ref)
    new_pct = np.histogram(new, breaks)[0] / len(new)

    psi_value = np.sum(sub_psi(ref_pct[i], new_pct[i]) for i in range(0, len(ref_pct)))
    return int(psi_value > psi_cutoff)

def get_data_drift (df_prev_compare, df_curr_compare):    
    dict_data_drift = dict()

    if isinstance(df_prev_compare , pd.Series):
        df_prev_compare   = df_prev_compare.to_frame()
    
    if isinstance(df_curr_compare , pd.Series):
        df_curr_compare   = df_curr_compare.to_frame()


    for c in df_curr_compare.columns:
        if pd.api.types.is_numeric_dtype(df_curr_compare[c]):                     
                drift = population_stability_index(df_prev_compare[c], df_curr_compare[c])
        else:
            drift = chi_square(df_prev_compare[c], df_curr_compare[c])
        dict_data_drift[c] = drift
    data_drift_flag = int(any(dict_data_drift.values()))    
    return  data_drift_flag, dict_data_drift

def get_column_residuals_outlier(df_curr_column, runtime):
    '''
    https://towardsdatascience.com/the-ultimate-guide-to-finding-outliers-in-your-time-series-data-part-1-1bf81e09ade4__;!!GMUSBZg!nwXhu_f2jwHgSIK7ElxsOicJKl6YEW0MYG1syVpMxOT1TIFdfHn2G7enUnyaJiZh7KxIeiqm9vGb4Ys$ 
    https://www.geeksforgeeks.org/seasonal-decomposition-of-time-series-by-loess-stl/
    compare new data (last entry) with the previous data residual assume they share the trend, sesonality
    '''  
    df_curr_column.index = pd.to_datetime(df_curr_column.index)
    df_curr_ind = df_curr_column[df_curr_column.index <= pd.to_datetime(runtime['train_end'])]
    # Perform STL decomposition
    stl = STL(df_curr_ind, seasonal=13, robust=False) #this exclude the last incomplete month
    result = stl.fit()
    #get the data up to second last
    residuals =  result.resid
    mean_resid = np.mean(residuals[:-1]) #this spare the last month
    std_resid = np.std(residuals[:-1])
    threshold = 3 * std_resid
    upper_limit = mean_resid + threshold
    lower_limit = mean_resid - threshold
    #compare last data with previous data
    outlier_flag = 0 if residuals[-1] <= upper_limit and residuals[-1] >= lower_limit else 1
    return outlier_flag, [residuals[-1], lower_limit, upper_limit] 

def get_outlier(df_curr, runtime):
    dict_outlier_flag = {}
    date_fields = ['trans_year','trans_month'] #Ignore these
    success_share_fields = ['success_share']
    for c in [i for i in df_curr.columns if i not in date_fields + success_share_fields ]:
        outlier_flag, detail = get_column_residuals_outlier(df_curr[c], runtime)
        dict_outlier_flag[c] = [outlier_flag, detail]
    if runtime['env']== 'dev':  #dev data in policy center s3 table is messy, almost each has outlier
        outlier_flag = dict_outlier_flag[next(iter(dict_outlier_flag))][0]   #this get the y outlier flag only, the data from database in the dictionary 
    else:  #if in prod or qa
        outlier_flag = int(any([v[0] for k, v in dict_outlier_flag.items() if k not in ['earned_premium_this_month']]))   #this metric has larger fluctuation, exclude it
    return outlier_flag,  dict_outlier_flag


def check_ts_missing_month(df, month_end):
    """check time series data if there is missing month"""
    month_begin = df.index.min().strftime("%Y-%m-%d")
    if not month_end:
        month_end = DtFind.first_day_last_month( dt.datetime.today(), dt_format=None).strftime("%Y-%m-%d")
    month_range =  pd.date_range(month_begin, month_end, 
                freq='MS').strftime("%Y-%m-%d").tolist()
    missing_month = [i for i in month_range if pd.to_datetime(i) not in df.index]
    return int(any(missing_month)), missing_month 

def convert_ts_to_dict (df):
    df.index = df.index.astype(str)
    return df.to_dict()


__all__ = ['get_best_prophet_param_both2']

from abc import ABC, abstractmethod
from cffds.tools._imports import *
from scipy.stats import uniform


class PerformanceTuneBase(ABC):


    def __init__(self, train_df, test_df, test_size, cf_type):
        self.train_df = train_df
        self.test_df = test_df        
        self.test_size = test_size
        self.cf_type = cf_type        
        self.column_mask_comb = get_column_boolean_combination(train_df, cf_type)    #at instance level
      

    @abstractmethod
    def objective(self):
        """Abstract method to create objective function."""
        pass
    

    def adjust_model_input(self, params): 
        '''
        this is to adjust prophet model input featuer per parameter value 
        '''           
        params_without_flag = {x:y for x, y in params.items() if x!='column_masks'}
        model = Prophet(**params_without_flag)
        column_masks = self.column_mask_comb[params['column_masks']]
        selected_columns = [ x  for x, y in zip(self.train_df.columns, column_masks) if y] 
        train_df = self.train_df.loc[:, selected_columns ]
        test_df = self.test_df.loc[:, selected_columns ]
        #add regressor
        additional_columns = selected_columns[2:]
        if additional_columns:
            for i in additional_columns:
                model.add_regressor(i)
        return model, train_df, test_df
    
    def get_additional_features(self, future, params, _train_df, _test_df ):
        column_masks =  self.column_mask_comb[params['column_masks']]               
        selected_columns = [ x for x, y in zip(_train_df.columns, column_masks) if y][2:] #skip ds and y
        for c in selected_columns:                 
            future[c] = pd.concat([_train_df[c], _test_df[c]])
        return future

class MangoTuning(PerformanceTuneBase):  
    
    conf_Dict = dict()
    conf_Dict['initial_random'] = 10
    conf_Dict['num_iteration'] = 100

    def __init__ (self, train_df, test_df, test_size, cf_type):
        super().__init__(train_df, test_df, test_size, cf_type)
        self.param_space = self.get_param_space()                   #using polymorphism for each child class

    def objective(self, args_list): 
        
        params_evaluated = []
        res = []
        
        for params in args_list:
            try:
                
                model, _train_df, _test_df = self.adjust_model_input(params)                
                model.fit(_train_df)                
                future = model.make_future_dataframe(periods=self.test_size, freq='MS') 
                # column_masks =  self.column_mask_comb[params['column_masks']]               
                # selected_columns = [ x for x, y in zip(_train_df.columns, column_masks) if y][2:] #skip ds and y
                # for c in selected_columns:                 
                #     future[c] = pd.concat([_train_df[c], _test_df[c]])
                future = self.get_additional_features(future, params, self.train_df, self.test_df)
                forecast = model.predict(future)
                predictions_tuned = forecast.tail(self.test_size)
                error = mape(_test_df['y'].values, predictions_tuned['yhat'].values)   
                
                params_evaluated.append(params)
                res.append(error)
            except Exception as e:            
                params_evaluated.append(params)
                print(e)
                res.append(100.0)# Giving high loss for exceptions regions of spaces
            
        return params_evaluated, res 
    
    def get_param_space(self):
         
         param_space = dict(growth = ['linear',  'flat'],  #remove 'logistic', need cap in data set
                   n_changepoints  = range(0, 55, 5),
                   changepoint_range  = uniform(0.5, 0.5),
                   yearly_seasonality = [True, False],
                   weekly_seasonality = [True, False],
                   daily_seasonality = [True, False],
                   seasonality_mode = ['additive', 'multiplicative'],
                   seasonality_prior_scale=uniform(5.0, 15.0),
                   changepoint_prior_scale=uniform(0.0, 0.1),
                   interval_width = uniform(0.2, 0.8),
                   uncertainty_samples = [500, 1000, 1500, 2000],
                   column_masks = range(0, len(self.column_mask_comb))   #direct use 
                   )        
         return param_space 
    
class OptunaTuning(PerformanceTuneBase):

    def __init__ (self, train_df, test_df, test_size, cf_type):
        super().__init__(train_df, test_df, test_size, cf_type)  

    def objective(self, trial): 
        params = {

            "growth":trial.suggest_categorical("growth", ['linear',  'flat']),  #remove 'logistic',  it will need cap
            "n_changepoints":trial.suggest_int("n_changepoints", 0, 55, 5),
            "changepoint_range":trial.suggest_float("changepoint_range", 0.5, 0.5),
            'yearly_seasonality': trial.suggest_categorical('yearly_seasonality',[True, False]),
            'weekly_seasonality': trial.suggest_categorical('weekly_seasonality',[True, False]),  
            'daily_seasonality': trial.suggest_categorical('daily_seasonality',[True, False]), 
            "seasonality_mode":trial.suggest_categorical("seasonality_mode", ['additive', 'multiplicative']),
            "seasonality_prior_scale":trial.suggest_float("seasonality_prior_scale", 5, 15),
            "changepoint_prior_scale":trial.suggest_float("changepoint_prior_scale", 0, 0.1),
            "interval_width":trial.suggest_float("interval_width", 0.2, 0.8),
            'uncertainty_samples': trial.suggest_categorical('uncertainty_samples',[500, 1000, 1500, 2000]), 
            'column_masks': trial.suggest_int("column_masks", 0, len(self.column_mask_comb)-1),  #optuna is inclusive for upper boundry, use -1
            
        }  
        
        try:
            model, _train_df, _test_df = self.adjust_model_input(params)        
            model.fit(_train_df)
            future = model.make_future_dataframe(periods=self.test_size, freq='MS')                 
            future = self.get_additional_features(future, params, self.train_df, self.test_df)
            forecast = model.predict(future)            
            predictions_tuned = forecast.tail(self.test_size)
            error = mape(_test_df['y'].values, predictions_tuned['yhat'].values)
            return error 
        except Exception as e:  
             print(e) 



def generate_boolean_combinations( n):
    """
    Generates a list of all possible boolean combinations for 'n' elements.

    Args:
        n: The number of boolean elements.

    Returns:
        A list of lists, where each inner list represents a unique boolean combination.
    """
    if n < 0:
        raise ValueError("Number of elements cannot be negative.")

    combinations = []
    for i in range(2**n):
        # Convert the integer 'i' to its binary representation
        # and pad with leading zeros to ensure 'n' digits.
        binary_string = bin(i)[2:].zfill(n)
        
        # Convert each binary digit to a boolean value
        boolean_list = [bool(int(bit)) for bit in binary_string]
        combinations.append(boolean_list)
    return combinations

def get_column_boolean_combination(df, cf_type):
    '''
    the prophet data format will be  [ds, y, additional columns]
    assume the first 2 columns are needed to predict, then from 3rd is optional
    '''        
    if df.shape[1] == 2:  #ds, y
        return [(True,True)]
    
    manditary_column_numb = 2
    if cf_type in ['salary', 'life_outflow_payroll',]:   # 'inflow_premiums' remove this for now
        manditary_column_numb = 3

    if df.shape[1] <2:
        raise ValueError("Dataframe must have at least two columns: 'ds' and 'y'.")
    elif df.shape[1]-manditary_column_numb <0:
        raise ValueError("Dataframe has less columns than required mandatory columns.")
    elif df.shape[1]-manditary_column_numb ==0:
        return [ [True]*df.shape[1] ]
    else:
        column_mask = generate_boolean_combinations(df.shape[1]-manditary_column_numb) #create combination except the first two columns
        
        #add first one in, need to convert between tuple and list    
        for i in  column_mask:
            i[0:0] = [True]*manditary_column_numb  #insert first two column mask
        #column_mask2 = [ tuple(i) for i in column_mask]   #remove tuple need 
        return column_mask

def mape(y_true, y_pred):
            return round(np.mean(np.abs((y_true - y_pred) / y_true)) * 100, 2)

def get_best_prophet_param_mango2 (train_df, test_df,  test_size, cf_type):
    mt = MangoTuning(train_df, test_df, test_size,cf_type)
    tuner = mango.Tuner(mt.param_space, mt.objective, mt.conf_Dict)
    results = tuner.minimize()
    best_params = results['best_params']
    best_params['column_masks'] =mt.column_mask_comb[best_params['column_masks']] #convert to boolean mask
    best_loss = results['best_objective']
    return best_params, best_loss  

def get_best_prophet_param_optuna2 (train_df, test_df,  test_size, cf_type):
    ot = OptunaTuning(train_df, test_df, test_size,cf_type)
    study = optuna.create_study(direction="minimize") 
    study.optimize(ot.objective, n_trials=100) # run the objective function 100 times
    optuna_best_params = study.best_params.copy()   #cannot assign value to study.best_params
    optuna_best_params['column_masks'] = ot.column_mask_comb[optuna_best_params['column_masks']] #convert to boolean mask       
    return optuna_best_params,study.best_value

def get_best_prophet_param_both2(train_df, test_df, test_size, cf_type):
    best_params_mango, best_loss_mango = get_best_prophet_param_mango2 (train_df, test_df,test_size, cf_type)
    best_params_optuna, best_loss_optuna = get_best_prophet_param_optuna2(train_df, test_df,test_size, cf_type)
    best_params = best_params_mango if best_loss_mango < best_loss_optuna else best_params_optuna
    best_params['columns_included'] = [ x  for x, y in zip(train_df.columns, best_params['column_masks'] ) if y]       
    best_loss = best_loss_mango if best_loss_mango < best_loss_optuna else best_loss_optuna  
    winner = 'mango' if best_loss_mango < best_loss_optuna else 'optuna' 
    results = {'mango':[best_loss_mango, best_params_mango], 'optuna':[best_loss_optuna,best_params_optuna]} 
    return best_params, best_loss, winner, results


from ._imports import *
from ._train import *
from ._pconfig import *
from ._misc import *
from ._monitor import monitor_pipeline
pd.set_option('display.float_format', lambda x: '%.2f' % x)

'''
cf_type = 'life_inflow_premiums'
#train_begin = None   
train_end = None 
env = 'dev'  
run_time = create_runtime (cf_type, env, train_end=None)
dict_results = training_pipeline( run_time)
df_score = get_forecast(dict_results, run_time)
df_score

'''


def scoring_pipeline (cf_type = 'claim loss', env = 'dev', logger = None, train_end = None):  
    '''
    Example
    -------
    This function will run monitor and training pipeline and use optmized model to forecast next three month
    example:
    scoring_pipeline (cf_type = 'claim loss', env = 'dev', logger = logger, train_end = '2024-10-01')
    add train_end to recreate previous month prediction
    loggile, logger = create_logger()
    scoring_pipeline (cf_type = 'claim loss', env = 'dev', logger = logger, train_end = None)

    Parameters
    ----------
    cf_type : str
        cash flow type.
    env : str
        environment that function is running.
    logger : logging.RootLogger
        keep job log.
    train_end : str
        training data end, by default previous month.

    Returns
    ------- 
    int
        Binary integer for error status.
    '''
    
    error_status = 0  
    try:    
        runtime = create_runtime(cf_type, env, train_end)
        if logger: 
            logger.info(f'''{cf_type} forecasting based on data between {runtime['train_begin']} and {runtime['train_end']},
                         test length is {runtime['test_len']} and forecast length is {runtime['forecast_len']}''')
        tracker = monitor_pipeline(runtime)        
        assert not tracker['Null Drift'], f'Data might be missing in {','.join(tracker['Missing Months'])} '        
        assert not tracker['Outlier Drift'], f'The latest record residue has 3 standard deviation from mean, please make sure data is right'
        dict_results = training_pipeline(runtime) 
        df_score = get_forecast(dict_results, runtime)
        #send to accounting
        if logger:logger.info(df_score.to_string())
        #send_sns(df_score , 'dl')  #TODO change dl to env when dev credential available        
        
        #save model asset
        save_model_assets(dict_results, runtime)
        
        #save monitor 
        save_monitor_mstr(tracker, runtime)
        #save_monitor_dashboard_value(runtime)
        
        #save score to ds folder
        save_cash_flow_score_to_ds(df_score, dict_results, runtime)        
        #export to silver layer
        save_cash_flow_score_to_silver (df_score, runtime)

    except Exception as e:        
        if logger:
            logger.error('Error', exc_info=e)
        error_status = 1
        raise  #pass error to outer try catch
    return error_status



#module variable, in case param cf_type different from folder name 
#life flow is not added, I will use standard name
dict_s3_prediction_folder_mapping = {
    'claim_loss': 'outflow_claims',
    'claim_subrogation': 'inflow_claims',
    'inflow_premiums':'inflow_premiums',
    'outflow_premiums':'outflow_premiums',
    'inflow_receipts': 'inflow_receipts', 
    'vendor':'outflow_vendor' ,
    'salary':'outflow_payroll'
    }

def create_runtime (cf_type, env, train_end=None):
    runtime = {}
    runtime['cf_type'] = cf_type.replace(' ', '_')
    
    if train_end:

        runtime['train_end'] = train_end
    else:         
         runtime['train_end'] = DtFind.first_day_last_month( dt.datetime.today(), dt_format=None).strftime("%Y-%m-%d")
        #runtime['train_end'] = '2024-10-01'
    month_offset = pconfig.get(f'{runtime['cf_type']}_month_offset', None)  #avoid create clutter record in pconfig
    if month_offset:
        runtime['train_begin'] = (dt.datetime.strptime(runtime['train_end'], "%Y-%m-%d") +pd.DateOffset(months=month_offset)).strftime("%Y-%m-%d")   #'2020-07-01'  from last month count back 52 month    
    else:
        runtime['train_begin'] = None  #if none, need to get that from query result    
    runtime['test_len'] = 8
    runtime['forecast_len'] = 3
    runtime['env'] = env    
    runtime['timestamp']      = dt.datetime.now().strftime("%Y%m%d_%H%M%S")
    return runtime


def select_data_range (df_ts, begin, end=None):  
    '''
    select_data_range(df_claim_ts, begin = '2020-04-01')
    '''
    if not begin:
        begin=df_ts.index[0]

    if not end:
        end = df_ts.index[-1]

    return df_ts.loc[begin:end]




############################ test new function 
def score_adjust_model_input(params, train_df, test_df): 
        '''
        this is to adjust prophet model input featuer per parameter value 
        '''           
        params_without_flag = {x:y for x, y in params.items() if x not in ['column_masks', 'columns_included']}
        model = Prophet(**params_without_flag)
        # column_masks = column_mask_comb[params['column_masks']]
        selected_columns = params['columns_included']
        train_df = train_df.loc[:, selected_columns ]
        test_df = test_df.loc[:, selected_columns ]
        #add regressor
        additional_columns = selected_columns[2:]
        if additional_columns:
            for i in additional_columns:
                model.add_regressor(i)
        return model, train_df, test_df
    
# def score_get_additional_features( future, params, _train_df, _test_df ):                 
#     additional_columns = params['columns_included'][2:] #skip ds and y
#     for c in additional_columns:                 
#         future[c] = pd.concat([_train_df[c], _test_df[c]])
#     return future

def get_forecast_month_features(future, runtime):
    if runtime['cf_type'] in ['outflow_premiums', 'inflow_premiums']:
        df_features = get_additional_features(runtime['env'])
        df_features.iloc[:, 0 ] = pd.to_datetime(df_features.iloc[:, 0 ])
        ts_column_name = df_features.columns.to_list()[0] 
        df_features.set_index(ts_column_name, drop=True, inplace=True)
        # inflow premimium has addition column from accounting 
        if runtime['cf_type'] == 'inflow_premiums':
            df_features0 = prep_premium_inflow_invoice(runtime['env'])   
            df_features0.index = (pd.to_datetime(df_features0[['expected_clear_year', 'expected_clear_month']]
                                                        .rename(columns={'expected_clear_year':'YEAR', 'expected_clear_month':'MONTH'})
                                                        .assign(DAY=1)))
            df_features = df_features0.join(df_features)
    elif runtime['cf_type'] in ['salary', 'life_outflow_payroll' ]:
        df_features = get_payday_datataframe()
        df_features['Date'] = pd.to_datetime(df_features['trans_year'].astype(str) + '-' + df_features['trans_month'] + '-01')
        df_features = df_features[['Date','payday_count','success_share']]
        df_features = df_features.set_index('Date')
    forecast_begin = future['ds'][-runtime['forecast_len']:].min()
    forecast_end = future['ds'][-runtime['forecast_len']:].max()
    filtered_features= select_data_range(df_features, forecast_begin, forecast_end)
    return filtered_features  

def get_forecast(dict_results, runtime):
    '''
    Use trained model to forecast for next three months 
    Parameters
    ----------
    dict_results : dict 
                a dictionary return from training step containing best hyperparm, best score as well data used for training.
    None : dict 
                a parameter set used in this process.
    Returns
    ----------
    None : DataFrame 
                prediction saved in  the dataframe .
    Examples
    ----------
    df_score = get_forecast(dict_results, runtime)
    
    '''
    df = dict_results['dataset inference']
    best_params = dict_results['best params']
    forecast_size = runtime['forecast_len']
    #modify model and df based on the best param, could 
    model, df, _ = score_adjust_model_input(best_params, df, df)  #use 2 df to meeting the orginal function requirement    
    model.fit(df)
    future = model.make_future_dataframe(periods=forecast_size, freq='MS')
    if len(best_params['columns_included'])>2:  #need to include other feature in future to forecast
            
            additional_features = get_forecast_month_features(future, runtime)
            column_name_list = df.iloc[:, 2:].columns.to_list()
            selected_columns =  [i for i in column_name_list if i in dict_results['best params'][ 'columns_included']] 
            for c in selected_columns:                 
                future[c] = pd.concat([df[c], additional_features[c]], ignore_index=True)
            dict_results['additional_feature_used'] = future
    forecast = model.predict(future)
    predictions_tuned = forecast.tail(forecast_size)
    dict_additional_columns ={
    "Type": dict_s3_prediction_folder_mapping.get(runtime['cf_type'], runtime['cf_type']).replace('_', ' '),
    "Timestamp" : str(runtime['timestamp']).replace('_', ' '),
    "Forecast Month": predictions_tuned['ds'].astype(str),
    "Values": predictions_tuned['yhat'],
    "Base Month":str(runtime['train_end']) }   
    predictions_tuned = predictions_tuned.assign(**dict_additional_columns)
    return predictions_tuned[['Timestamp', 'Type', 'Forecast Month', 'Values', 'Base Month']] 

#### save model results
def save_model_assets (dict_results,runtime):
    '''
    this function saves model asset to AWS S3 
    Parameters
    ----------
    dict_results : dict 
                dictionary contains all training meta data.
    runtime : dict 
                dictionary contains all parameter used for training  and scoring.
    Examples
    ----------
    save_model_assets (dict_results,runtime)
    '''
    #first time run, if no version file, create one    
    if not (version:= get_model_version(runtime)):
           archive_model_version (runtime)

    #only update version when end date == last month, to avoid testing update
    if runtime['train_end'] ==DtFind.first_day_last_month( dt.datetime.today(), dt_format=None).strftime("%Y-%m-%d"):
             archive_model_version (runtime)
    
    #handle model with more than 1 feature
    #dict_results_without_flag = {k:v for k, v in dict_results['best params'].items() if k != 'additional_feature_flag'}

    params_without_flag = {x:y for x, y in dict_results['best params'].items() if x not in ['column_masks', 'columns_included']}
        

    #save model performance 
    s3_save_and_upload_file(
            #get best params and best loss
            {k: dict_results[k] for k in ('best params', 'best loss', 'tuning results')},
            s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{runtime['env']}'], f'cash_flow_model_{runtime["cf_type"]}_{runtime["timestamp"]}.json'),
            overwrite='overwrite',
            env=runtime['env'],
        )

    #save model itself
    s3_save_and_upload_file(
                
        Prophet(**params_without_flag),
        s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{runtime['env']}'], f'cash_flow_model_{runtime["cf_type"]}_{runtime["timestamp"]}.pkl'),
        overwrite='overwrite',
        env=runtime['env'],
    )

    #save model ref dataset
    s3_save_and_upload_file(
        dict_results['original dataset'],
        s3_file_path = os.path.join(pconfig[f's3_model_assets_folder_{runtime['env']}'], f'cash_flow_model_{runtime["cf_type"]}_{runtime["timestamp"]}.csv'),
        overwrite='overwrite',
        env=runtime['env'],
    )

#save score data
def save_cash_flow_score_to_ds(df_score, dict_results, runtime):
    '''
    Save inference dataset, additional features and prediction result in s3 for future audition 
    Parameters
    ----------
    df_score : DataFrame 
                dataFrame containing prediction.
    dict_results : dict 
                dictionary cotaining all training metadata.
    runtime : dict 
                dictionary contain  training params.
    Examples
    ----------
    save_cash_flow_score_to_ds(df_score, dict_results, runtime)
    '''
    s3_save_and_upload_file(
        dict_results['dataset inference'],
        s3_file_path = os.path.join(pconfig[f's3_model_output_eds_folder_{runtime['env']}'], f'cash_flow_model_{runtime["cf_type"]}_inference_data_{runtime["timestamp"]}.csv'),
        overwrite='overwrite',
        env=runtime['env'],
    )

    if 'additional_feature_used' in dict_results:
        s3_save_and_upload_file(
        dict_results['additional_feature_used'],
        s3_file_path = os.path.join(pconfig[f's3_model_output_eds_folder_{runtime['env']}'], f'cash_flow_model_{runtime["cf_type"]}_additional_features_{runtime["timestamp"]}.csv'),
        overwrite='overwrite',
        env=runtime['env'],
    )

    s3_save_and_upload_file(
        df_score,
        s3_file_path = os.path.join(pconfig[f's3_model_output_mstr_folder_{runtime['env']}'], f'score/cash_flow_model_{runtime["cf_type"]}_scoring_{runtime["timestamp"]}.csv'),
        overwrite='overwrite',
        env=runtime['env'],
    )


# save prediction to silver
def save_cash_flow_score_to_silver (df_score, runtime): 
    '''
    Save score to cashflow_forecast/forecast_stage/ 
    Parameters
    ----------
    df_score : DataFrame 
                dataframe contains score results.
    runtime : dict 
                dictionary contains training params.
    Examples
    ----------
    save_cash_flow_score_to_silver (df_score, runtime)
    '''   
    accounting_type = dict_s3_prediction_folder_mapping.get(runtime['cf_type'], runtime['cf_type']) #if cannot find, use type value
    run_date = runtime['timestamp'].split('_')[0]
    type = runtime['cf_type']
    base_month = runtime['train_end'][:7]
    file_name = f'{accounting_type}_{base_month}_{run_date}.parquet'
    local_file_path = config['local']['local_folder']
    local_file_path = os.path.join(local_file_path, file_name)
    s3_file_path = f'cashflow_forecast/forecast_stage/{accounting_type}/{file_name}'
    df_score.to_parquet(local_file_path)
    s3_upload_file_to_silver(local_file_path, s3_file_path, overwrite=True, env=runtime['env'])    
    os.remove(local_file_path)
    return  

def save_monitor_mstr(data, runtime):
    '''
    Save model data monitoring result to s3 folder 
    Parameters
    ----------
    data : dict 
                dictionary that conatins satistics comparing current data vs previous run.
    runtime : dict 
                dictionary contains training params info.
    Examples
    ----------
    save_monitor_mstr(tracker, runtime)
    '''
    if not runtime["train_begin"]:        #some model don't have fixed start date 
         keyname = next(iter(data['Curr Dataset']))
         ts = list(data['Curr Dataset'][keyname].keys())
         runtime["train_begin"] = min(ts)
         
    s3_save_and_upload_json(
        data,
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_mstr_folder_{runtime["env"]}'],
            'monitor/',
            f'{runtime['cf_type']}_monitor_{runtime["train_begin"]}_{runtime["train_end"]}_{runtime["timestamp"]}.json'
        ),
        env=runtime['env'],
    )    
    return



#### these function are  for model data monitor by wcao 
def get_latest_monitor_file(env = 'dev'):
    monitor_files = s3_list_folder_contents(
                os.path.join(pconfig[f's3_model_output_mstr_folder_{env}'], 'monitor/'),
                pattern=r'^.*\.json$',
                env=env,
            )
    df_monitor_files = pd.DataFrame(columns=['file_path'])
    df_monitor_files['file_path'] = monitor_files
    df_monitor_files['file_name'] = df_monitor_files['file_path'].str.split('/').str[-1]
    df_monitor_files['cash_flow_type'] = df_monitor_files['file_name'].str.split('_').str[:-5].apply(lambda x: "_".join(x))
    df_monitor_files['training_end_month'] = df_monitor_files['file_name'].str.split('_').str[-3]
    df_monitor_files['training_date_time'] = df_monitor_files['file_name'].str.split('_').str[-2:].apply(lambda x: "_".join(x)).str.strip('.json')
    #drop duplicate
    df_monitor_files2 = df_monitor_files.sort_values(by=['cash_flow_type','training_end_month', 'training_date_time'])
    df_monitor_files2 = df_monitor_files2.drop_duplicates(subset=['cash_flow_type','training_end_month'], keep='last', ignore_index=True ) 
    return df_monitor_files2

def extract_monitor_value(df_monitor, env):
    df_monitor_extract = pd.DataFrame(columns = ['type', 'train_end',  'run_timestamp', 'data_drift',
                                              'outlier_drift', 'null_drift', 'residuals', 'low_limit', 'high_limit', 'mean_limit'])
    for i, row in df_monitor.iterrows():
        dict_row = s3_download_and_read_json(row['file_path'],None, None, env)
        if type(dict_row['Outlier Detail']) == list:  #this is old format
            Outlier_Detail= dict_row['Outlier Detail']
        elif type(dict_row['Outlier Detail']) == dict:
            Outlier_Detail=  next(iter(dict_row['Outlier Detail'].values()), None)[1]  #get first value
        residuals, low_limit, high_limit,  = Outlier_Detail
        mean_limit = (low_limit + high_limit)/2
        df_monitor_extract.loc[len(df_monitor_extract)] =[row['cash_flow_type'],row['training_end_month'], row['training_date_time'], dict_row['Data Drift'],
                                                        dict_row['Outlier Drift'], dict_row['Null Drift'],residuals, low_limit, high_limit, mean_limit]
    dict_type_map = {
        'claim_loss': 'outflow claims', 'claim_subrogation':'inflow claims',  
        'salary':'outflow payroll',  'vendor':'outflow vendor' 
    }
    df_monitor_extract['type'] =  df_monitor_extract['type'].apply(lambda x:dict_type_map.get(x, x.replace('_', " ")))
    return df_monitor_extract 

def save_monitor_dashboard_value (env):
    '''
    collect latest monitor file and extract the value for monitor dashboard use 
    Parameters
    ----------
    env : str 
                environment the function is running.
    Examples
    ----------
    save_monitor_dashboard_value('dev')

    '''  
    df_monitor = get_latest_monitor_file(env)
    df_monitor_extract = extract_monitor_value(df_monitor, env)
    s3_save_and_upload_csv(
        df_monitor_extract,
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_mstr_folder_{env}'] + r'dashboard/',           
            f'data_monitor.csv'
        ),
        overwrite=True,
        env=env
    )
    return

####these function is for combine prediction into one file by wcao

def get_latest_prediction_file(env = 'dev'):
    file_list = s3_get_prediction_file_in_silver (s3_folder_path = r"cashflow_forecast/forecast_stage/", s3_bucket = f'amica-aws-eds-{env}-dl-silver', pattern = r'^.*\.parquet$', env=env)
    df_prediction_file = pd.DataFrame()
    df_prediction_file['file_path'] = file_list
    df_prediction_file['timestamp'] =  df_prediction_file['file_path'].str.split('/').str[-1].str.split('_').str[-1]
    df_prediction_file['base_month'] = df_prediction_file['file_path'].str.split('/').str[-1].str.split('_').str[-2]
    df_prediction_file['category'] = df_prediction_file['file_path'].str.split('/').str[-1].str.split('_').str[:-2].apply(lambda x:'_'.join(x))
    #drop duplicate
    df_prediction_file2 =  df_prediction_file.sort_values(by=['category','base_month', 'timestamp'])
    df_prediction_file2 = df_prediction_file2.drop_duplicates(subset=['category','base_month'], keep='last', ignore_index=True ) 
    return df_prediction_file2 

def extract_prediction(df_prediction_file, env = 'dev' ): 
    #some old file contain old name passing by param, this mapping fix, in new env no this issue
    dict_category_mapping  = { x.replace('_', " "):y.replace('_', " ")  for x, y in dict_s3_prediction_folder_mapping.items()  }
       
    df_prediction =  pd.DataFrame(columns=['Timestamp', 'Type', 'Forecast Month', 'Values', 'Base Month'])
    for i, row in df_prediction_file.iterrows():
        df_file = s3_download_and_read_file(row['file_path'], local_file_path=None, s3_bucket=f'amica-aws-eds-{env}-dl-silver', read_file_kwargs={}, env=env)
        df_file['Type']= df_file['Type'].apply(lambda x: dict_category_mapping.get(x, x))  #this fix old name to new name
        df_prediction = pd.concat([df_prediction, df_file], ignore_index= True)
        #remove local file after load
        s3_remove_temp_local_download (row['file_path'])
    df_prediction = df_prediction.sort_values(by=['Type', 'Base Month', 'Timestamp'], ignore_index= True)
    return df_prediction

def save_latest_prediction (env): 
    '''
    collect latest model prediction from different cash flow type to prediction_combined.csv 
    Parameters
    ----------
    env : str 
                environment the function is running.
    Examples
    ----------
    save_latest_prediction('dev')
    ''' 
    df_prediction_file = get_latest_prediction_file(env = env)
    df_prediction_extract = extract_prediction(df_prediction_file, env = env )
    timestamp =  df_prediction_extract['Timestamp'].max().replace(' ', '_')
    local_file_path = os.path.join(config['local']['local_folder'], 'prediction_combined.csv' )    
    df_prediction_extract.to_csv(local_file_path, index = False)
    s3_file_path = r'cashflow_forecast/forecast_stage/forecast_combined/' + 'prediction_combined_' + timestamp + '.csv'
    s3_upload_file_to_silver(local_file_path, s3_file_path, overwrite=True, env=env)
    os.remove(local_file_path)    
    return 




from setuptools import find_namespace_packages, setup, find_packages

setup(name='eds_ds_tools',
      version='1.3',
      description='tools for data science',
      #packages=['eds_ds_tools'],
      author_email='wenleicao@gmail.com',
      zip_safe=True,
       include_package_data=True,      
      #packages=find_packages(where="."),
      # packages=find_namespace_packages(where="."),
      #   package_dir={"": "."},
      #   package_data={
      #       'eds_ds_tools': ['*.zip'], # Include all .zip files in 'eds_ds_tools'
      #   },
      #packages = find_packages(where="."),
      packages= ['eds_ds_tools', 'scripts'],
      
      
      )



__all__ = ['evaluate_pipeline']

from ._imports import *
from ._pconfig import *
from ._train import dataset_target, evaluate_metrics
from ._misc import *	
from ._features import feature_cancel_renewal, feature_cancel_expiring

def evaluate_pipeline(run_start, run_end, env='dev', logger=None, slicer=None):
    """
    Evaluation pipeline that loads past scores, pulls target within time period, 
    and evaluate performance for binary classification.

    Parameters
    ----------
    run_start : str
        The start month string in '%Y-%m' format.
    run_end : str
        The end month string in '%Y-%m' format.
    env : str
        The run environment, 'dev' or 'prod'.
    logger : logging.RootLogger
        Logger for information, warnings, and errors.
    slicer : function
        Function to divides data into subgroup for evaluation.
        Default (None) evaluates on the entire dataset.
        Example: slicer = lambda r: 'Over 60' if r['Age']>60 else 'Under 60'

    Returns
    -------
    int
        Binary integer for error status.
    """

    error_status = 0
    trackers = {}
    impacts  = {}
    if logger:
        logger.info(f'Evaluate from {run_start} to {run_end} on {env}.')
    try:
        runtime   = create_evaluate_runtime(run_start, run_end, env, logger)
        df_scores = dataset_load_scores(runtime)
        if len(df_scores)==0:
            raise('No Records to Evaluate')
        df_target = dataset_target_initialize(df_scores, runtime)
        df_target = dataset_target(df_target, runtime)
        df        = pd.merge(df_scores, df_target[['POLICY_KEY', pconfig['target']]], how='left', on=['POLICY_KEY'])

        evals    = evaluate_metrics(df[pconfig['target']], df['PROBABILITY'], df['PREDICTION'])
        trackers = {'all': evals}
        if slicer:
            evals_slicer = evaluate_metrics_by_slice(df, slicer)
            trackers     = trackers | evals_slicer
        trackers['error'] = error_status

        impacts = impact_metrics(df, runtime)

    except Exception as e:
        if logger:
            logger.error('Error', exc_info=e)
        error_status = 1
        trackers     = {'error': error_status}
    finally:
        save_evals_mstr(trackers, runtime)
        save_impact_mstr(impacts, runtime)
        return error_status

def create_evaluate_runtime(run_start, run_end, env, logger):
    runtime                   = {}
    runtime['timestamp']      = dt.datetime.now().strftime("%Y%m%d%H%M%S")
    runtime['env']            = env
    runtime['logger']         = logger
    runtime['teradata_table'] = f'{config["teradata"]["user"]}.temp_table'
    runtime['genesys_table']  = '#tmp_phone_number'
    runtime['type_dt']        = pconfig['type_dt']

    run_start = dt.datetime.strptime(run_start, '%Y-%m')
    run_end   = dt.datetime.strptime(run_end, '%Y-%m')
    run_date  = DtFind.first_day_last_month(run_start).replace(day=pconfig['run_day'])
    runtime['run_start']      = DtFind.first_day_of_month(run_start).strftime('%Y-%m-%d')
    runtime['run_end']        = DtFind.last_day_of_month(run_end).strftime('%Y-%m-%d')

    runtime['run_date']       = run_date.strftime('%Y-%m-%d')
    runtime['run_cutoff']     = DtFind.first_day_of_month(run_date + dt.timedelta(3 * 30)).replace(day=pconfig['run_day']).strftime('%Y-%m-%d')

    runtime['features']       =  {'CANCEL_RENEWAL': True, 'CANCEL_EXPIRING': True}
    return runtime  

def evaluate_metrics_by_slice(df, slicer):
    trackers = {}
    slices = df.apply(slicer, axis=1)
    for s in slices.unique():
        df_slice    = df.loc[slices==s]
        evals       = evaluate_metrics(df_slice[pconfig['target']], df_slice['PROBABILITY'], df_slice['PREDICTION'])
        trackers[s] = evals
    return trackers

def impact_metrics(df, runtime):
    #df_care    = dataset_phone_care(df.loc[df['PREDICTION']==1], runtime)
    df_care    = dataset_phone_care_not_available (df)
    df_genesys = dataset_phone_genesys(df.loc[df['PREDICTION']==1], runtime)
    df_impact  = pd.concat([
        df_care.loc[(df_care['PREDICTION']==1) & (df_care[pconfig['target']]==0) & (df_care['calldurationinseconds']>0), ['ACCOUNT_NO','PREMIUM_RENEWAL']],
        df_genesys.loc[(df_genesys['PREDICTION']==1) & (df_genesys[pconfig['target']]==0) & (df_genesys['SegmentTime']>0), ['ACCOUNT_NO','PREMIUM_RENEWAL']],
    ], ignore_index=True).drop_duplicates()

    impact_care = {
        'accounts predicted cancel':          df_care['ACCOUNT_NO'].nunique(),
        'accounts predicted cancel & called': df_care.dropna(subset=['calldurationinseconds'])['ACCOUNT_NO'].nunique(),
        'calls total':                        int(df_care['calldurationinseconds'].count()),
        'calls inbound':                      int(df_care.apply(lambda x: x['calldurationinseconds'] if x['calltype']=='Inbound' else None, axis=1).count()),
        'calls outbound':                     int(df_care.apply(lambda x: x['calldurationinseconds'] if x['calltype']=='Outbound' else None, axis =1).count()),
        'average call duration (sec)':        np.round(df_care['calldurationinseconds'].mean(), 2),
    }
    impact_genesys = {
        'accounts predicted cancel':          df_genesys['ACCOUNT_NO'].nunique(),
        'accounts predicted cancel & called': df_genesys.dropna(subset=['SegmentTime'])['ACCOUNT_NO'].nunique(),
        'calls total':                        int(df_genesys['SegmentTime'].count()),
        'calls inbound':                      int(df_genesys.apply(lambda x: x['SegmentTime'] if x['direction']=='inbound' else None, axis=1).count()),
        'calls outbound':                     int(df_genesys.apply(lambda x: x['SegmentTime'] if x['direction']=='outbound' else None, axis=1).count()),        
        'average call duration (sec)':        np.round(df_genesys['SegmentTime'].mean(), 2),
    }

    impact_retained = {
        'premium gained':    int(df_impact['PREMIUM_RENEWAL'].sum()),
        'accounts retained': len(df_impact), 
    }

    impacts = {
        'care':     impact_care,
        'genesys':  impact_genesys,
        'retained': impact_retained,
    }

    return impacts

def dataset_load_scores(runtime):
    fnames = s3_list_folder_contents(pconfig[f's3_model_output_eds_folder_{runtime["env"]}'], env=runtime['env'])
    fnames = pd.DataFrame({'file': fnames})
    pat    = r'cancel_([\d\-]*)_([\d\-]*)_([\d]*)\.csv'
    fnames[['run_month', 'run_date', 'timestamp']] = fnames['file'].astype(str).str.extract(pat)
    fnames = fnames.loc[
        (fnames['run_month'] >= runtime['run_start']) 
        &
        (fnames['run_month'] <= runtime['run_end'])
    ]
    
    df = pd.DataFrame()
    for _, row in fnames.iterrows():
        fname  = row['file']
        ts     = row['timestamp']
        df_run = s3_download_and_read_csv(fname, env=runtime['env'])
        df_run = df_run.assign(timestamp = ts)
        df = pd.concat([df, df_run], ignore_index=True)
    df = df.sort_values(by='timestamp', ascending=False).drop_duplicates(subset='POLICY_KEY', keep='first')
    df['POLICY_KEY'] = df['POLICY_KEY'].astype(str).str.strip()
    return df

def dataset_target_initialize(df, runtime):
    table_name  = runtime['teradata_table']
    key_list    = ','.join([k.strip() for k in df['POLICY_KEY'].astype(str).tolist()])   

    table_query = f'''
        select
            trim(a.policy_key) as POLICY_KEY
            , trim(a.policy_no) as POLICY_NO	
            , trim(a.account_no) as ACCOUNT_NO	
            , trim(a.perpetual_pol_no) as PERPETUAL_POL_NO	
            , trim(a.expired_policy_no) as EXPIRED_POLICY_NO	
            , a.effective_date as EFFECTIVE_DATE	
            , a.expiration_date as EXPIRATION_DATE	
            , a.process_ts as PROCESS_TS 
        from atuprv.dw_agreement a
        where
            a.policy_type = 1
            and a.policy_activity = 'Renewal'
            and a.policy_key in ({key_list})
        qualify row_number() over (partition by a.policy_key order by a.process_ts desc)=1
    '''
    td_drop_table(table_name)
    td_volatile(table_query, table_name)
    df = td_query(query_creator(table_columns='*', table_name=table_name, table_ops=''))
    df['POLICY_KEY'] = df['POLICY_KEY'].astype(str).str.strip()

    if runtime['features']['CANCEL_RENEWAL']:
        df = feature_cancel_renewal(df, runtime)	
    if runtime['features']['CANCEL_EXPIRING']:
        df = feature_cancel_expiring (df, runtime)	
    return df

def dataset_phone_care(df, runtime):
    table_columns = '''
        pol.name as policy_number,
        FinServ__Account__c, 
        acct.name, 
        acct.personcontactid, 
        task.subject, 
        task.callType, 
        task.callobject, 
        task.calldisposition, 
        task.calldurationinseconds,
        task.createddate
    '''
    table_name = 'salesforce.finserv__alert__c alert'
    table_ops= f'''
        inner join salesforce.account acct on alert.finserv__account__c = acct.sfid
        inner join salesforce.task task on task.whoid = acct.personcontactid
        inner join salesforce.insurancepolicy pol on pol.nameinsuredid = alert.FinServ__Account__c and task.createddate between pol.effectivedate and pol.expirationdate  --eliminate 2 
        where finserv__message__c ='Retention Risk' 
        and pol.effectivedate between '{runtime['run_start']}' and '{runtime['run_end']}'
        and task.callType is not null
        and pol.line_of_business_type__c = 'Private Passenger'
        and task.createddate between '{runtime['run_date']}' and '{runtime['run_cutoff']}'
    '''
    table_query = query_creator(table_columns, table_name, table_ops)
    df_query = sf_query(table_query)

    df = df.merge(df_query, how='left', left_on='POLICY_NO', right_on='policy_number')
    return df

def dataset_phone_genesys(df, runtime):
    table_columns = '''
        agree.policy_key, agree.policy_no, trim(agree.account_no) as account_no, agrol.edb_id, pm.phone_id,
            trim('tel:+1' || phonenum.area_code || phonenum.exchange || phonenum.number_rw) as TelephoneNumber
    '''
    table_name = '''
        (
            select policy_key, policy_no, account_no, process_ts from atuprv.cust_dw_agreement_latest where policy_type=1
            qualify row_number() over (partition by account_no order by process_ts desc)=1 
        ) agree
    '''
    table_ops= f'''
        left join (
            select policy_key, edb_id from atuprv_adhoc.dw_e_agree_role_latest 
            qualify row_number() over (partition by policy_key order by seq_id)=1 
        ) agrol on agree.policy_key=agrol.policy_key
        left join (
            select edb_id, phone_id from atuprv.dw_e_party_phone
            qualify row_number() over (partition by edb_id order by process_ts desc)=1
        ) pm on agrol.edb_id = pm.edb_id
        left join atuprv.dw_e_phone phonenum on phonenum.phone_id = pm.phone_id    
        where year(agree.process_ts) > year('{runtime['run_cutoff']}')-5
    '''
    table_query = query_creator(table_columns, table_name, table_ops)
    df_mapping_query = td_query(table_query)
    df_genesys = df.merge(df_mapping_query[['account_no','TelephoneNumber']] , how='left', left_on='ACCOUNT_NO', right_on='account_no')

    gs_temp(df_genesys[['TelephoneNumber']].drop_duplicates(), runtime['genesys_table'], columns_index=['TelephoneNumber'])
    

    #optimize the query, add index 
    genesys_call_index_sql = f"""
    drop table if exists #tmp_call 

    create table #tmp_call (
    phone  varchar(300),
    userid varchar(50),
    direction varchar(10),
    SegmentTime numeric
    )
    
    insert into #tmp_call 
    select 
    iif( a.direction = 'inbound',  a.Ani, a.Dnis) as phone,
    a.USERID,
                a.direction as direction,
                round(a.tTalkcomplete/1000, 2) as SegmentTime        

    from callsegments_emite a
    where 
        a.ConversationStart between '{runtime['run_date']}' and '{runtime['run_cutoff']}'
    and a.tTalkcomplete is not null

    create index idx_phone on #tmp_call(phone) 

    select * from #tmp_call

    """
    gs_query(genesys_call_index_sql)

    genesys_phone_sql =  f"""
        select    
            a.phone,
            a.direction as direction,
            a.SegmentTime        
        from #tmp_call a
        join {runtime['genesys_table']} b on  b.TelephoneNumber  = a.phone
        join dbo.genesysCloudUsersAPI c on a.USERID = c.[entity_id]
        where 
             c.division_name like 'SCS%'
        """
    
    df_phone_query = gs_query(genesys_phone_sql)

    df_genesys = df_genesys.merge(df_phone_query[['phone', 'SegmentTime', 'direction']] , how='left', left_on='TelephoneNumber', right_on='phone')
    return df_genesys

def save_evals_mstr(data, runtime):
    s3_save_and_upload_json(
        data,
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_mstr_folder_{runtime["env"]}'],
            'evaluate/',
            f'cancel_eval_{runtime["run_start"]}_{runtime["run_end"]}_{runtime["timestamp"]}.json'
        ),
        env=runtime['env'],
    )

def save_impact_mstr(data, runtime):
    s3_save_and_upload_json(
        data,
        s3_file_path = os.path.join(
            pconfig[f's3_model_output_mstr_folder_{runtime["env"]}'],
            'impact/',
            f'cancel_impact_{runtime["run_start"]}_{runtime["run_end"]}_{runtime["timestamp"]}.json'
        ),
        env=runtime['env'],
    )

def dataset_phone_care_not_available (df):
    '''
    this function is used to replace the orginal df when 
    care data source is not available, so pipeline still run
    '''
    df_care = pd.DataFrame().reindex(columns=list(df.columns) + ['policy_number', 'finserv__account__c', 'name', 'personcontactid',
       'subject', 'calltype', 'callobject', 'calldisposition',
       'calldurationinseconds', 'createddate'])
    return df_care